﻿Type=Service
Version=4.3
ModulesStructureVersion=1
B4A=true
@EndOfDesignText@
#Region  Service Attributes 
	#StartAtBoot: False
#End Region
	
' GLOBAL EVENTS -----------------------------------------------------------------------------------------
	'#Event: ServiceStarted
' -------------------------------------------------------------------------------------------------------

' SERVER MODE EVENTS ------------------------------------------------------------------------------------
	'#Event: ServerStarted(ServerIP as string)
	'#Event: ServerStopped
	'#Event: ClientConnected(GUID as string)
	'#Event: ClientConnectError(ErrorMessage as string)
 	'#Event: AStreamError(GUID as string, ErrorMessage as string)
	'#Event: DataFromClient(GUID as string, Data as string)
	'#Event: ClientDisconnected(GUID as string)
' -------------------------------------------------------------------------------------------------------

' CLIENT MODE EVENTS ------------------------------------------------------------------------------------
	'#Event: ConnectedToServer
	'#Event: DisconnectedFromServer
	'#Event: ServerConnectError(ErrorMessage as string)
	'#Event: DataFromServer(Data as string)
 	'#Event: ClientAStreamError(ErrorMessage as string)
' -------------------------------------------------------------------------------------------------------





Sub Process_Globals
	' Constants
	Public Const SERVER_PORT As Int = 50505
	Public ServiceStarted As Boolean = False 

	' Types (Structures)
	Type ConnectionInfo(oSocket As Socket, GUID As String, LastContact As Long)
	
	' Local variables
	Private moServer As ServerSocket		' socket to listen for connections
	Private moClientSock As Socket 			' socket to use for client mode conection
	Private moClientAStream As AsyncStreams ' Async Stream to use for client mode connection
	Private moConnections As Map  			' connection collection: key = the async stream for the socket, value = ConnectionInfo type
	Private msServerIP As String			' IP address of the server
    
	Private Const DEBUG_MODE As Boolean = True 
	Private moCallBack As Object
	Private msEvent As String 
	Private mbServer As Boolean = False
	Private mbClient As Boolean = False 
	Private mbSending As Boolean = False 

	Private tmrPing As Timer 
	
End Sub


#Region "  Service Code (general service routines) "
Private Sub Service_Create
	MyLog("Service_Create")
	
	' initial server mode connection list
	moConnections.Initialize 
	' fill in the IP of this machine	
    Dim poServer As ServerSocket
	poServer.Initialize(SERVER_PORT, "")
	msServerIP = poServer.GetMyIP 
	poServer.Close
End Sub

Private Sub Service_Start (StartingIntent As Intent)
	MyLog("Service_Start")
	ServiceStarted = True 
	tmrPing.Initialize("tmrPing", 6000)
	tmrPing.Enabled = True
End Sub

Private Sub Service_Destroy
	tmrPing.Enabled = False 
	ServiceStarted = False
	MyLog("Service_Destroy")
	If mbServer Then
		MyLog("  ServerMode stopping...")
		moServer.Close 
		RaiseEvent("ServerStopped", Null)
	    If moConnections.IsInitialized Then
			Dim piIndex As Int
			For piIndex = moConnections.Size - 1 To 0 Step -1
				Dim poAStream As AsyncStreams 
				Dim poConn As ConnectionInfo 
				poAStream = moConnections.GetKeyAt(piIndex)
				poConn = moConnections.GetValueAt(piIndex)
				If poAStream.IsInitialized Then
		    		MyLog("    Async Streams Closed for " & poConn.GUID)
		    		poAStream.Close
				End If
				If poConn.IsInitialized Then
					If poConn.oSocket.IsInitialized Then
						If poConn.oSocket.Connected Then
						    MyLog("    Connection " & poConn.GUID & " closed.")
							poConn.oSocket.Close 
						End If
					End If
				End If 
				moConnections.Remove(poAStream)
			Next
		End If
	End If 
	If mbClient Then
		Client_Disconnect
	End If 
End Sub

'   Connection sanity check, polls all connections and disconnects if not active.  For some reason the asyncstreams_Terminated event never fires.
Private Sub tmrPing_Tick
	Dim piIndex As Int 
	
	If mbServer AND moConnections.IsInitialized Then
		For piIndex = moConnections.Size - 1 To 0 Step -1
			Dim poConn As ConnectionInfo 
			poConn = moConnections.GetValueAt(piIndex)
			If DateTime.Now - poConn.LastContact > 5500 Then
				Disconnect_Client(poConn.GUID)
			End If
		Next
	Else
		If mbClient AND moClientSock.IsInitialized Then
			If mbSending = False Then
				If moClientSock.Connected Then
					Send_ToServer(Chr(46))
				Else
					Client_Disconnect
				End If
			End If 
		End If
	End If 
End Sub 



#End Region


#Region "  Server Mode Code "
'  Puts the service into <B>SERVER MODE</b> (listen for connections from clients).  Call this after starting service.
Public Sub Start_Server(oCallBack As Object, EventName As String)
	If mbClient Then
		Return
	End If 
	MyLog("Start_Server Mode")
	
	If moConnections.IsInitialized = False Then
		moConnections.Initialize 
	Else
		moConnections.Clear 
	End If 
	MyLog("    Connection pool initialized.")

	
	moCallBack = oCallBack
	msEvent = EventName & "_"

	MyLog("    Callback and event set.")

	moServer.Initialize(SERVER_PORT, "Server")
    moServer.Listen

	MyLog("    Server listen socket initialized.")

	' raise event that we're running	
	RaiseEvent("ServerStarted", Null)
	mbServer = True

	MyLog("    ServerStarted event raised.")
End Sub 

'  Removes the service from <b>SERVER MODE</b>.  You can start client mode after this call.
Public Sub Stop_Server
	If mbServer = False Then
		Return
	End If
	MyLog("Stop_Server")
	moServer.Close 
	Disconnect_Client("")
	RaiseEvent("ServerStopped", Null)
	mbServer = False
End Sub 

Private Sub Server_NewConnection (Successful As Boolean, NewSocket As Socket)
	Dim poList As List
	poList.Initialize

	MyLog("Server_NewConnection")
	If Successful Then
		Dim poNew As ConnectionInfo
		poNew.Initialize 
		poNew.GUID = Generate_GUID
		poNew.oSocket = NewSocket
		poNew.LastContact = DateTime.Now 
		Dim poAStream As AsyncStreams 
		poAStream.InitializePrefix(NewSocket.InputStream, False, NewSocket.OutputStream, "AStreams")
		moConnections.Put(poAStream, poNew)
		poList.Add(poNew.GUID)
		RaiseEvent("ClientConnected", poList)
		MyLog("    Successful connection, new GUID = " & poNew.GUID)
    Else
		MyLog("    Connection Failed: " & LastException.Message)
		poList.Add(LastException.Message)
		RaiseEvent("ClientConnectError", poList)
    End If
    moServer.Listen
End Sub

Private Sub AStreams_NewData (Buffer() As Byte)
	Dim poList As List
	Dim poConn As ConnectionInfo 
	Dim psMsg As String

	MyLog("AStreams_NewData")
	' update last contact time 
	poConn = moConnections.Get(Sender)
	poConn.LastContact = DateTime.Now 
	moConnections.Put(Sender, poConn)
	If Buffer.Length = 1 Then
		If Buffer(0) = 46 Then
			MyLog("    Sanity check, discarded")
			' sanity check, discard
			Return
		End If
	End If 
	' fire event
	poList.Initialize 
	poList.Add(poConn.GUID)
    psMsg = BytesToString(Buffer, 0, Buffer.Length, "UTF8")
	poList.Add(psMsg)
	MyLog("    " & psMsg.Length & " bytes: " & psMsg)
	RaiseEvent("DataFromClient", poList)
End Sub

'  Raised when the client has terminted the connection
Private Sub AStreams_Terminated
	Dim poList As List
	Dim poConn As ConnectionInfo 
	Dim poAStream As AsyncStreams 

	MyLog("AStreams_Terminated")
	poAStream = Sender
	poList.Initialize 
	poConn = moConnections.Get(Sender)
	poList.Add(poConn.GUID)
	poAStream.Close 
	poConn.oSocket.Close 
	moConnections.Remove(Sender)
	MyLog("    Lost client: " & poConn.GUID) 
	RaiseEvent("ClientDisconnected", poList)
End Sub

Private Sub AStreams_Error
	Dim poList As List
	Dim poConn As ConnectionInfo 
	MyLog("AStreams_Error")
	poList.Initialize 
	poConn = moConnections.Get(Sender)
	poList.Add(poConn.GUID)
	poList.Add(LastException.Message)
	MyLog("    GUID: " & poConn.GUID) 
	MyLog("    ERROR: " & LastException.Message) 
	RaiseEvent("AStreamError", poList)
End Sub

'  Send data to connnected client(s).  Set GUID to "" to send the data to all clients, or to a specific GUID for just one client.
Public Sub Send_ToClient(GUID As String, Data As String)
	Dim piIndex As Int
	
	MyLog("Send_ToClient")

	mbSending = True
	
	For piIndex = 0 To moConnections.Size - 1
		Dim poConn As ConnectionInfo 
		Dim poAStream As AsyncStreams 
		
		poAStream = moConnections.GetKeyAt(piIndex)
		If poAStream.IsInitialized Then
			poConn = moConnections.GetValueAt(piIndex)
			If poConn.IsInitialized Then
				If poConn.GUID = GUID OR GUID = "" Then
					If poConn.oSocket.Connected Then
						poAStream.Write(Data.GetBytes("UTF8"))
						MyLog("    GUID: " & poConn.GUID & " / Wrote " & Data.Length & " bytes: " & Data)
					Else
						MyLog("    GUID: " & poConn.GUID & " / No write, socket NOT connected.")
						Disconnect_Client(poConn.GUID)
					End If
					If GUID <> "" Then
						Exit
					End If 
				End If
			End If
		End If
	Next 
	mbSending = False 
End Sub 

Public Sub Send_ToAllClients(ExceptionGUID As String, Data As String)
	Dim piIndex As Int
	
	MyLog("Send_ToAllClients")
	
	mbSending = True 
	For piIndex = 0 To moConnections.Size - 1
		Dim poConn As ConnectionInfo 
		Dim poAStream As AsyncStreams 
		
		poAStream = moConnections.GetKeyAt(piIndex)
		If poAStream.IsInitialized Then
			poConn = moConnections.GetValueAt(piIndex)
			If poConn.IsInitialized Then
				If poConn.GUID <> ExceptionGUID Then
					If poConn.oSocket.Connected Then
						poAStream.Write(Data.GetBytes("UTF8"))
						MyLog("    GUID: " & poConn.GUID & " / Wrote " & Data.Length & " bytes: " & Data)
					Else
						MyLog("    GUID: " & poConn.GUID & " / No write, socket NOT connected.")
						Disconnect_Client(poConn.GUID)
					End If
				End If
			End If
		End If
	Next 
	mbSending = False 
End Sub 

'  Supply the GUID of the client you wish to disconnect, or supply "" to disconnect <B>ALL</B> clients.
Public Sub Disconnect_Client(GUID As String)
	Dim piIndex As Int
	Dim poList As List
	
	MyLog("Disconnect_Client")

	poList.Initialize
	poList.Add("")
	
	For piIndex = moConnections.Size - 1 To 0 Step -1
		Dim poConn As ConnectionInfo 
		Dim poAStream As AsyncStreams 
		
		poAStream = moConnections.GetKeyAt(piIndex)
		poConn = moConnections.GetValueAt(piIndex)
		If poConn.GUID = GUID OR GUID = "" Then
			If poConn.oSocket.Connected Then
				poAStream.Close 
				poConn.oSocket.Close 
				poList.Set(0, poConn.GUID)
				MyLog("    GUID: " & poConn.GUID & " Disconnected.")
				RaiseEvent("ClientDisconnected", poList)
			End If
			moConnections.Remove(poAStream)
		End If
	Next 
End Sub 


'  Returns the number of <b>CONNECTED</b> clients.
Public Sub Num_Connections As Int
	Dim piIndex As Int
	Dim piCount As Int = 0 
	
	For piIndex = 0 To moConnections.Size - 1
		Dim poConn As ConnectionInfo 
		poConn = moConnections.GetValueAt(piIndex)
		If poConn.IsInitialized Then
			If poConn.oSocket.Connected Then
				piCount = piCount + 1
			End If
		End If
	Next 
	
	Return piCount
End Sub 

'  Returns the IP address assigned to this device
Public Sub Get_ServerIP As String
	Return msServerIP
End Sub 

'  Returns the IP address of the requested client connection
Public Sub Get_ClientIP(ClientGUID As String) As String
	Dim psIP As String = ""
	Dim piIndex As Int
	
	If moConnections.IsInitialized Then
		For piIndex = 0 To moConnections.Size - 1
			Dim poConn As ConnectionInfo 
			poConn = moConnections.GetValueAt(piIndex)
			If ClientGUID = poConn.GUID Then
				Dim r As Reflector
   				r.Target = poConn.oSocket 
   				r.Target = r.GetField("socket")
   				r.Target = r.RunMethod("getInetAddress") 'InetAddress
   				psIP = r.RunMethod("getHostAddress")
				Exit
			End If
		Next
	End If
	Return psIP
End Sub 


#End Region


'====================================================================================
'####################################################################################
'====================================================================================


#Region "  Client Mode Code "
'  Puts the service into <b>CLIENT MODE</b>.  Call this after starting service.
Public Sub Start_ClientMode(oCallBack As Object, EventName As String)
	If mbServer Then 
		Return
	End If
	
	MyLog("Start_ClientMode")
	moCallBack = oCallBack
	msEvent = EventName & "_"
	
	Client_Disconnect
	mbClient = True
End Sub

' Removes the service from <b>CLIENT MODE</b>.  You may start server mode after calling this function.
Public Sub Stop_ClientMode
	If mbClient Then
		Client_Disconnect
		mbClient = False
	End If 
End Sub 

'  Returns TRUE if in <b>CLIENT MODE</b> and service is ready to connect to a server.
Public Sub Client_Ready As Boolean
	Return mbClient
End Sub 

'  Connects this device to the requested IP address.  Service must be in client mode before this call.
Public Sub Client_Connect(DestinationIP As String)
	Client_Disconnect
	MyLog("Client_Connect: Host(" & DestinationIP & ") Port(" & SERVER_PORT & ")")
	If moClientSock.IsInitialized = False Then
		moClientSock.Initialize("ClientSock")
	End If
	moClientSock.Connect(DestinationIP, SERVER_PORT, 15000)
End Sub 

'  Returns TRUE if client mode is connected to a server.  Otherwise false.	
Public Sub Connected_ToServer() As Boolean
	If moClientSock.IsInitialized Then
		If moClientSock.Connected Then
			Return True
		End If
	End If
	Return False 
End Sub 

Private Sub ClientSock_Connected (Successful As Boolean)
	MyLog("ClientSock_Connected")
	If Successful Then
		MyLog("   Connected to server.")
		moClientAStream.InitializePrefix(moClientSock.InputStream, False, moClientSock.OutputStream, "ClientAStream")
		RaiseEvent("ConnectedToServer", Null)
	Else
		MyLog("   Error connecting to server: " & LastException.Message)
		moClientSock.Close 
		Dim poList As List
		poList.Initialize 
		poList.Add(LastException.Message)
		RaiseEvent("ServerConnectError", poList)
	End If 
End Sub	
	
Private Sub ClientAStream_NewData (Buffer() As Byte)
	Dim poList As List

	MyLog("ClientAstream_NewData")
	MyLog("    " & Buffer.Length & " bytes received.")
	If Buffer.Length = 1 Then
		If Buffer(0) = 46 Then
			MyLog("    Sanity check, discarded")
			' sanity check, discard
			Return
		End If
	End If 
	poList.Initialize 
	poList.Add(BytesToString(Buffer, 0, Buffer.Length, "UTF8"))
	MyLog("    Data Recevied: " & poList.Get(0))
	' let user know we've received data	
	RaiseEvent("DataFromServer", poList)
End Sub

'  Raised when the server has terminted the connection
Private Sub ClientAStream_Terminated
	MyLog("ClientAStream_Terminated")
	' make sure we're disconnected	
	Client_Disconnect 
End Sub

Private Sub ClientAStream_Error
	Dim poList As List

	MyLog("ClientAStream_Error")
	MyLog("    ERROR: " & LastException.Message) 
	poList.Initialize 
	poList.Add(LastException.Message)
	RaiseEvent("ClientAStreamError", poList)
	' make sure we're disconnected	
	Client_Disconnect
End Sub

'  Call to disconnect client from server when in client mode.  Client mode will still be active after this call, only Stop_ClientMode will end client mode.
Public Sub Client_Disconnect
	If mbClient Then
		MyLog("Client_Disconnect")
		If moClientAStream.IsInitialized Then
			moClientAStream.Close 
		End If
		If moClientSock.IsInitialized Then
			If moClientSock.Connected Then
				moClientSock.Close 
			End If
		End If
		RaiseEvent("DisconnectedFromServer", Null)
	End If 
End Sub 	

'  Sends data from client to server when in <b>CLIENT MODE</b>.
Public Sub Send_ToServer(Data As String)
	If mbClient Then
		mbSending = True 
		MyLog("Send_ToServer")
		If moClientAStream.IsInitialized Then
			If moClientSock.IsInitialized Then
				If moClientSock.Connected Then
					moClientAStream.Write(Data.GetBytes("UTF8"))
					MyLog("    Sent " & Data.Length & " bytes to server: " & Data)
				Else
					MyLog("    ERROR, client not connected to server.")
					Client_Disconnect
				End If
			End If
		End If
		mbSending = False 
	End If
End Sub 





#End Region




#Region "  Private support modules "

Private Sub Generate_GUID As String
	Dim r As Reflector
 	r.Target = r.RunStaticMethod("java.util.UUID", "randomUUID", Null, Null)
 	Return r.RunMethod("toString")
End Sub 

Private Sub RaiseEvent(EventName As String, EventParms As List)
	If moCallBack <> Null Then
		If SubExists(moCallBack, msEvent & EventName) Then
			Select Case EventName

				' SERVER MODE EVENTS
				Case "ServerStarted"
					CallSubDelayed2(moCallBack, msEvent & EventName, msServerIP)

				Case "ServerStopped"
					CallSubDelayed(moCallBack, msEvent & EventName)

				Case "ClientConnected"
					CallSubDelayed2(moCallBack, msEvent & EventName, EventParms.Get(0))							' 0 = GUID

				Case "ClientConnectError"
					CallSubDelayed2(moCallBack, msEvent & EventName, EventParms.Get(0))							' 0 = Connect Error Message

				Case "AStreamError"
					CallSubDelayed3(moCallBack, msEvent & EventName, EventParms.Get(0), EventParms.Get(1)) 		' 0 = GUID, 1 = error message
				
				Case "DataFromClient"
					CallSubDelayed3(moCallBack, msEvent & EventName, EventParms.Get(0), EventParms.Get(1))		' 0 = GUID, 1 = data received
				
				Case "ClientDisconnected"
					CallSubDelayed2(moCallBack, msEvent & EventName, EventParms.Get(0))							' 0 = GUID
					

				' CLIENT MODE EVENTS					
				Case "DisconnectedFromServer"					
					CallSubDelayed(moCallBack, msEvent & EventName)

				Case "ServerConnectError"
					CallSubDelayed2(moCallBack, msEvent & EventName, EventParms.Get(0))							' 0 = connection failure error message
					
				Case "ConnectedToServer"
					CallSubDelayed(moCallBack, msEvent & EventName)
					
				Case "DataFromServer"
					CallSubDelayed2(moCallBack, msEvent & EventName, EventParms.Get(0))							' 0 = data received from the server
					
				Case "ClientAStreamError"
					CallSubDelayed2(moCallBack, msEvent & EventName, EventParms.Get(0))							' 0 = astream error message
				
			End Select
		End If
	End If
End Sub 

Private Sub MyLog(Message As String)
	If DEBUG_MODE Then
	    LogColor("srvSockets - " & Message, Colors.RGB(50, 200, 50))
	End If
End Sub 


#End Region





